# Deep Learning Containers in HPC 🧠📦

Author: [Hauke Kirchner](mailto:hauke.kirchner@gwdg.de)

Reliable software environments are often a problem for data scientists. This becomes an issue especially when the analysis needs to be run on different machines - colleagues, cloud or HPC cluster. Here we introduce the idea of containerised software environments in the use case of a deep learning workflow. A more general view on this has already been described in our blog post [Declutter your Python environments](https://gitlab-ce.gwdg.de/hpc-team-public/science-domains-blog/-/blob/main/20230907_python-apptainer.md).

<figure>
  <img
    src="./img/dlc.png"
    alt="deep learning container">
  <figcaption>
  Overview of a deep learning container. This example deep learning container may be a bit overloaded, but it shows the flexibility and versatility you can get when using containers to wrap your software environment.
  </figcaption>
</figure>


## Why? Lets ask [Chat AI](https://chat-ai.academiccloud.de/) 💡

> prompt: What are the five most important benefits of using containers in my machine learning workflow?

1. **Consistency and Reproducibility**: Containers ensure that your ML environment is consistent across different machines and environments, making it easier to reproduce results and collaborate with others.
2. **Dependency Management**: Containers simplify dependency management by packaging all required libraries and dependencies within the container, eliminating version conflicts and ensuring that your ML code runs smoothly.
3. **Isolation and Security**: Containers provide a sandboxed environment for your ML code, isolating it from the host system and preventing potential security risks, such as data breaches or malware infections.
4. **Portability and Flexibility**: Containers are highly portable, allowing you to move your ML workloads between different environments, such as development, testing, and production, without worrying about compatibility issues.
5. **Efficient Resource Utilization**: Containers can run multiple ML workloads on a single machine, maximizing resource utilization and reducing the need for additional hardware.

Credits: Here the self-hosted model LLaMA 3 70B Instruct was used to generate this answer. Only an excerpt of the original answer is shown here.

## Building the container 🛠️

```
me@local-machine~ % ssh glogin
[me@glogin9 ~]$ cd /path/to/deep-learning-with-gpu-cores/code/container
[me@glogin9 ~]$ bash build_dlc-dlwgpu.sh
```

The container can be built on our various HPC clusters using the `build_dlc-*.sh` scripts. Details on how to use Apptainer and the process of building containers are described in our [documentation](https://docs.hpc.gwdg.de/software_stacks/list_of_modules/apptainer/index.html) and our blog article [Decluttering your Python environments](https://gitlab-ce.gwdg.de/hpc-team-public/science-domains-blog/-/blob/main/20230907_python-apptainer.md).
Running `build_dlc-dlwgpu.sh` will build an image with the software used for our workshop example, as defined in `dlc-dlwgpu.def`. Contrary to the traditional way of using conda to install all packages defined in a `requirements.txt` file, pip is used here to reduce the number of software packages used. However, there are good reasons to use conda, so `build_dlc-conda-example.sh` shows a minimal example of installing Python packages in a container using conda (see `dlc-conda-example.def`).

If you encounter errors while building the container, have a look at `build_dlc-dlwgpu.log` or `build_dlc-conda-example.log`. You can also use `cat` (in a different terminal) to see the progress of the image building process.

The image can now be used on our JupyterHub service or via the command line on our HPC systems. We recommend using JupyterHub for the initial phase of a project to set up the workflow and for data exploration. Compute-intensive workloads, such as model training, should be run on the HPC cluster using SLURM. This allows users to execute computational jobs in a very flexible and customisable way, making the best use of the available computing resources.

💎 Building a container can take some time. So we provide an image at the path given below so that you can continue to explore the potential of containers on our cluster: `/scratch/projects/workshops/gpu-workshop/dlc-dlwgpu.sif`

## Using the container on HPC JupyterHub 🎨

Below are two examples of how to use HPC JupyterHub. The first example shows the use of a JupyterNotebook for data exploration. The second example is about testing the software environment to ensure that everything is set up correctly, e.g. CUDA for GPU acceleration.

<figure>
  <img
    src="./img/hpc-jupyterhub-screenshot.png"
    alt="Screenshot of the HPC JupyterHub with examples">
  <figcaption>
  Screenshot of the HPC JupyterHub with examples
  </figcaption>
</figure>

1. visit https://jupyter.hpc.gwdg.de
2. click on the "Sign in with AcademicCloud" button
3. start your server with the following options (everything else is default)
  - Select the appropriate HPC project (username)
	- Select a job profile: `GWDG HPC with GPU
  - Choose "Custom Apptainer Container Location"
	- Set your own Apptainer container location: `path/to/dlc-dlwgpu.sif`
	- Jupyter Notebook's Home directory: `/path/to/deep-learning-with-gpu-cores/`
4. run the code
    - Notebook: Open `/path/to/deep-learning-with-gpu-cores/code/notebooks/00.1-inital-data-exploration.ipynb`
    - Python file: 
        - Open `/path/to/deep-learning-with-gpu-cores/test_env.py`.
        - [Create Console for Editor](https://jupyterlab.readthedocs.io/en/stable/user/documents_kernels.html) and run the code in the console with `Shift + Enter`.
5. Stop the JupyterHub server
	- File -> Hub Control Panel -> Stop My Server 

💎 `cat ~/current.jupyterhub.notebook.log` can give you more details on your JupyterHub sever.

## Using the container via CLI on the HPC 🏭

The full potential of an HPC cluster can only be utilised using the command line interface (CLI). Workflows can be optimized for the available hardware, such as different accelerators (e.g. GPUs, NPU, ...) and highly parallel workflows. Here a simple workflow using our workshops example is shown, based on the same container, that was used in JupyterHub. For more details please have a look at our [documentation](https://docs.hpc.gwdg.de/index.html), e.g. on [Slurm/GPU Usage](https://docs.hpc.gwdg.de/how_to_use/slurm/gpu_usage/index.html) and [GPU Partitions](https://docs.hpc.gwdg.de/how_to_use/compute_partitions/gpu_partitions/index.html).

### Interactive HPC usage

Using an HPC cluster interactively gives you direct access to the cluster's compute resources. This can be particularly useful for optimising your workflow on the HPC. For the interactive usage first the compute ressources are requested (`salloc`). On the requested compute node, a shell is started within the container, allowing code to be executed with the pre-built software environment. In this example, the software environment is tested to make sure that everything works as intended, such as the correct use of CUDA for GPU acceleration (`test_env.py`).

```bash
cd path/to/deep-learning-with-gpu-cores/code/

# get available GPUs with (see https://docs.hpc.gwdg.de/how_to_use/slurm/gpu_usage/index.html)
# sinfo -o "%25N  %5c  %10m  %32f  %10G %18P " | grep gpu

# grete
salloc -t 01:00:00 -p grete:interactive -N1 -G 3g.20gb

# kisski
salloc -t 01:00:00 -p kisski -N1 -G A100:1

module load apptainer

apptainer exec --nv --bind /scratch /path/to/dlc-dlwgpu.sif bash

python test_env.py

# exit container
exit

#exit salloc
exit
```

### HPC usage via batch scripts

Finally, the defined workflow can be submitted as a job to the workload manager [SLURM](https://docs.hpc.gwdg.de/how_to_use/slurm/index.html). To do this, a job script needs to be defined and submitted via `sbatch`. As you already have a lot of experience with the containerised software environment, starting on your local machine and JupyterHub, a major point of failure when scaling up your analysis and moving to HPC systems has become unlikely. 

```bash
sbatch path/to/deep-learning-with-gpu-cores/code/submit_train_dlc.sh
```

## Hands on

### 1 Reproduce the given examples.

Here you should get some experience with building containers on the cluster, and how to use the containerized software in your workflows on JupyterHub and the HPC. All the relevant code, commands and settings are given above in this README. 

### 2 Adapt the container to one of your own use cases.

The real power of containers is that you can define your own personal set of software that you can take with you wherever you need it. So let's start with the interesting part and build your own custom containers that fit your needs. Below we have prepared some examples that have worked for us in various projects to give you some inspiration. 

```
mkdir ~/tmp && cd $_
cp ~/deep-learning-with-gpu-cores/code/container/{dlc-conda-example.def,build_dlc-conda-example.sh} .
```

You can install RStudio with custom R packages or you chosse to install Julia. Just extend the `%post` section int the definition file. Here you can find some inspiration, but feel encourged to try it with your set of relevant software.

1. Have a look at `dlc-python-r.def`

2. Julia example

```
    # Install Julia
    apt install -y julia

    echo 'ENV["HTTP_PROXY"] = "http://www-cache.gwdg.de:3128"' >> /etc/julia/startup.jl
    echo 'ENV["HTTPS_PROXY"] = "http://www-cache.gwdg.de:3128"' >> /etc/julia/startup.jl
```

The software is now available on the HPC via your own container, but can also be used in our [JupyterHub service](https://jupyter.hpc.gwdg.de/).

### 3 **Advanced**: Try to build and/or run containers on your local machine.

So far we have limited ourselves to the use of containers on our GWDG systems/HPC. You have seen that the whole workflow, from container building to interactive use via JupyterHub to SLURM-based job execution using all the potential of HPC, can be implemented based on our services. However, a big selling point of containers is that you can take them with you. So let us try to build and/or run containers on your local machine. As this is highly dependent on your operating system, we will not provide step-by-step instructions, you will have to figure it out for yourself. In this way you will become more familiar with containers and how to use them on different systems. The Science Domains blog post [Declutter your Python environments](https://gitlab-ce.gwdg.de/hpc-team-public/science-domains-blog/-/blob/main/20230907_python-apptainer.md) gives an example for [Linux systems](https://gitlab-ce.gwdg.de/hpc-team-public/science-domains-blog/-/blob/main/20230907_python-apptainer.md?ref_type=heads#i-want-to-test-it-on-my-own-system). In addition, [UTM](https://mac.getutm.app/) can be a good choice and has been tested for MacOS.